# 陈一潇
# 1400017639

# 实现UnorderedList的如下方法：
# append, index, pop, insert
# 用于列表字符串表示的__str__方法
# 用于取切片的__getslice__方法，其定义同python列表lst[i:j]


class Node:
    """节点类，其中把返回数据，返回尾部设为了属性，并定义了str方法，其余与教材相同。"""

    def __init__(self, init_data):
        self.__data = init_data
        self.__next = None

    def __str__(self):
        return str((self.__data, self.__next))

    @property
    def data(self):
        return self.__data

    @property
    def next(self):
        return self.__next

    def set_data(self, new_data):
        self.__data = new_data

    def set_next(self, new_next):
        self.__next = new_next


class UnorderedList:
    """无序表，元素按照插入的顺序排列，定义了除已有方法外，定义了append，index，insert，pop以及字符串和访问和切片的方法。
    将已有方法有所重写，定义了长度变量和tail变量，降低了一些方法的时间复杂度。遗憾的是，pop()方法的时间复杂度依旧是O(n)。
    包含了一些简单的异常处理。"""

    def __init__(self):
        self.__head = None
        self.__tail = None
        self.__len = 0

    def isEmpty(self):
        return not self.__head

    def add(self, item):
        self.__len += 1
        temp = Node(item)
        temp.set_next(self.__head)
        if not self.__head:
            self.__head = self.__tail = temp
        else:
            self.__head = temp

    def __len__(self):
        return self.__len

    def size(self):
        return self.__len

    def search(self, item):
        current = self.__head
        while current:
            if current.data == item:
                return True
            else:
                current = current.next
        return False

    def remove(self, item):
        """为原方法增加了若数据不在表中则报错的设计。"""
        current = self.__head
        previous = None
        found = False
        while not found:
            if current.data == item:
                found = True
            else:
                previous = current
                current = current.next

        if not previous:
            self.__head = current.next
        else:
            previous.set_next(current.next)
        if not previous.next:
            self.__tail = previous

        if not found:
            raise ValueError("{} is not in list".format(item))
        else:
            self.__len -= 1

    def append(self, item):
        """从列表尾部加入一个元素。"""
        self.__len += 1
        temp = Node(item)
        if not self.__head:
            self.__head = self.__tail = temp
        else:
            self.__tail.set_next(temp)
            self.__tail = temp

    def index(self, item):
        """返回item元素所对应的索引值，若item不在表中，则报错。"""
        current = self.__head
        found = False
        loc = 0
        while not found:
            if current.data == item:
                found = True
            else:
                current = current.next
                loc += 1
        if found:
            return loc
        else:
            raise ValueError("{} is not in list".format(item))

    def insert(self, pos, item):
        """将数据项插入到位置pos，pos必须是一个整数，否则报错，若为负数将从后往前数。
        要求原列表具有足够多个数据项，能让item占据位置pos，否则默认插入在最后一个。"""
        assert isinstance(pos, int)
        if pos < 0:
            pos = self.__len + pos
        pos = min(pos, self.__len)

        temp = Node(item)
        current = self.__head
        if pos == 0:
            temp.set_next(current)
        elif pos == self.__len:
            self.__tail.set_next(temp)
            self.__tail = temp
        else:
            for i in range(pos-1):
                current = current.next
            temp.set_next(current.next)
            current.set_next(temp)

        self.__len += 1

    def pop(self, pos=None):
        """返回指定位置pos的数据项，并将其从列表中移除。若不指定则默认为最后一个。
        负数同样为从后往前数，若不是整数或超出范围则会报错。"""
        if pos == None:
            pos = self.__len - 1

        assert isinstance(pos, int) and pos in range(-self.__len, self.__len)
        if pos < 0:
            pos = self.__len + pos
        self.__len -= 1

        current = self.__head
        if pos == 0:
            self.__head = current.next
            return current.data
        else:
            for i in range(pos-1):
                current = current.next
            temp = current.next
            current.set_next(temp.next)
            return temp.data


    def __str__(self):
        """用于列表字符串表示的方法"""
        current = self.__head
        lst = []
        while current:
            lst.append(current.data)
            current = current.next
        return str(lst)

    def __getitem__(self, item):
        """访问元素或切片，使用方式与列表的完全相同，支持step。
        而且支持负数哦……就是可能存在一些未知的bug。"""
        if isinstance(item, int):
            if item < 0:
                item = self.__len + item
            if item not in range(self.__len):
                raise IndexError("list index out of range")
            current = self.__head
            for i in range(item):
                current = current.next
            return current.data
        elif isinstance(item, slice):
            start = item.start if item.start > 0 else self.__len + item.start
            stop = item.stop if item.stop > 0 else self.__len + item.stop
            current = self.__head
            lst = []
            for i in range(stop):
                r = range(start, stop, item.step) if item.step else range(start, stop)
                if i in r:
                    lst.append(current.data)
                current = current.next
            return lst
        else:
            raise TypeError("list indices must be integers or slices, not str")


if __name__ == '__main__':  # 测试接口，随便写了一些东西，大致反映了定义的方法的功能。

    mylist = UnorderedList()

    mylist.add(31)
    mylist.add(77)
    mylist.add(17)
    mylist.add(93)
    mylist.add(26)
    mylist.add(54)
    mylist.add(100)
    print(mylist.search(100))
    print(len(mylist))
    print(mylist)

    mylist.remove(54)
    mylist.remove(93)
    mylist.remove(31)
    print(mylist.size())
    print(mylist.search(93))
    print(mylist)

    mylist.append(3)
    print(mylist.index(3))
    print(mylist.index(100))
    print(mylist)

    mylist.insert(-1, 250)
    print(mylist.index(250))
    print(mylist)
    print(mylist.pop())
    print(mylist)

    print(mylist[0])
    print(mylist[1:-1:2])